@@ -14,16 +14,17 @@
 
 
 from functools import partial
-from collections import Counter
 
 import numpy as onp
 
 from jax import core
-from jax.util import unzip2, prod
+from jax.util import unzip2
 from jax.tree_util import (register_pytree_node, tree_structure,
                            treedef_is_leaf, tree_flatten, tree_unflatten)
 import jax.linear_util as lu
-
+from jax.interpreters import xla
+from jax.lax import lax
+from jax.lax import lax_fft
 
 def jet(fun, primals, series):
   try:
@@ -48,19 +49,32 @@ def jet(fun, primals, series):
     yield tree_flatten(ans)
 
   f, out_tree = flatten_fun_output(lu.wrap_init(fun))
-  out_primals, out_terms = jet_transform(f).call_wrapped(primals, series)
+  out_primals, out_terms = jet_fun(jet_subtrace(f)).call_wrapped(primals, series)
   return tree_unflatten(out_tree(), out_primals), tree_unflatten(out_tree(), out_terms)
 
 @lu.transformation
-def jet_transform(primals, series):
+def jet_fun(primals, series):
   with core.new_master(JetTrace) as master:
-    trace = JetTrace(master, core.cur_sublevel())
-    in_tracers = map(partial(JetTracer, trace), primals, series)
-    ans = yield in_tracers, {}
-    out_tracers = map(trace.full_raise, ans)
-    out_primals, out_terms = unzip2((t.primal, t.terms) for t in out_tracers)
+    out_primals, out_terms = yield (master, primals, series), {}
+    del master
   yield out_primals, out_terms
 
+@lu.transformation
+def jet_subtrace(master, primals, series):
+  trace = JetTrace(master, core.cur_sublevel())
+  in_tracers = map(partial(JetTracer, trace), primals, series)
+  ans = yield in_tracers, {}
+  out_tracers = map(trace.full_raise, ans)
+  out_primals, out_terms = unzip2((t.primal, t.terms) for t in out_tracers)
+  yield out_primals, out_terms
+
+@lu.transformation_with_aux
+def traceable(in_tree_def, *primals_and_series):
+  primals_in, series_in = tree_unflatten(in_tree_def, primals_and_series)
+  primals_out, series_out = yield (primals_in, series_in), {}
+  out_flat, out_tree_def = tree_flatten((primals_out, series_out))
+  yield out_flat, out_tree_def
+
 
 class JetTracer(core.Tracer):
   __slots__ = ["primal", "terms"]
@@ -93,6 +107,7 @@ class JetTrace(core.Trace):
     return JetTracer(self, val.primal, val.terms)
 
   def process_primitive(self, primitive, tracers, params):
+    assert not primitive.multiple_results  # TODO
     primals_in, series_in = unzip2((t.primal, t.terms) for t in tracers)
     order, = {len(terms) for terms in series_in if terms is not zero_series}
     series_in = [[zero_term] * order if s is zero_series else s
@@ -106,10 +121,23 @@ class JetTrace(core.Trace):
     return JetTracer(self, primal_out, terms_out)
 
   def process_call(self, call_primitive, f, tracers, params):
-    assert False  # TODO
+    primals_in, series_in = unzip2((t.primal, t.terms) for t in tracers)
+    primals_and_series, in_tree_def = tree_flatten((primals_in, series_in))
+    f_jet, out_tree_def = traceable(jet_subtrace(f, self.master), in_tree_def)
+    result = call_primitive.bind(f_jet, *primals_and_series, **params)
+    primals_out, series_out = tree_unflatten(out_tree_def(), result)
+    return [JetTracer(self, p, ts) for p, ts in zip(primals_out, series_out)]
 
-  def post_process_call(self, call_primitive, out_tracer, params):
-    assert False  # TODO
+  def post_process_call(self, call_primitive, out_tracers, params):
+    primals, series = unzip2((t.primal, t.terms) for t in out_tracers)
+    out, treedef = tree_flatten((primals, series))
+    del primals, series
+    master = self.master
+    def todo(x):
+      primals, series = tree_unflatten(treedef, x)
+      trace = JetTrace(master, core.cur_sublevel())
+      return map(partial(JetTracer, trace), primals, series)
+    return out, todo
 
   def join(self, xt, yt):
     assert False  # TODO?
@@ -124,8 +152,33 @@ zero_series = ZeroSeries()
 register_pytree_node(ZeroSeries, lambda z: ((), None), lambda _, xs: zero_series)
 
 
+### rule definitions
+
 jet_rules = {}
 
+def defzero(prim):
+  jet_rules[prim] = partial(zero_prop, prim)
+
+def zero_prop(prim, primals_in, series_in, **params):
+  primal_out = prim.bind(*primals_in, **params)
+  return primal_out, zero_series
+
+defzero(lax.le_p)
+defzero(lax.lt_p)
+defzero(lax.gt_p)
+defzero(lax.ge_p)
+defzero(lax.eq_p)
+defzero(lax.ne_p)
+defzero(lax.and_p)
+defzero(lax.or_p)
+defzero(lax.xor_p)
+defzero(lax.floor_p)
+defzero(lax.ceil_p)
+defzero(lax.round_p)
+defzero(lax.sign_p)
+defzero(lax.stop_gradient_p)
+
+
 def deflinear(prim):
   jet_rules[prim] = partial(linear_prop, prim)
 
@@ -134,14 +187,6 @@ def linear_prop(prim, primals_in, series_in, **params):
   series_out = [prim.bind(*terms_in, **params) for terms_in in zip(*series_in)]
   return primal_out, series_out
 
-
-### rule definitions
-
-from jax.lax import lax
-
-def fact(n):
-  return lax.exp(lax.lgamma(n+1.))
-
 deflinear(lax.neg_p)
 deflinear(lax.real_p)
 deflinear(lax.complex_p)
@@ -159,32 +204,65 @@ deflinear(lax.slice_p)
 deflinear(lax.reduce_sum_p)
 deflinear(lax.reduce_window_sum_p)
 deflinear(lax.tie_in_p)
+deflinear(lax_fft.fft_p)
+deflinear(xla.device_put_p)
+
+
+
+### More complicated rules
+
+def fact(n):
+  return lax.exp(lax.lgamma(n+1.))
+
+def _scale(k, j):
+  return 1. / (fact(k - j) * fact(j - 1))
 
 def _exp_taylor(primals_in, series_in):
   x, = primals_in
   series, = series_in
   u = [x] + series
   v = [lax.exp(x)] + [None] * len(series)
-  def scale(k, j): return 1. / (fact(k-j) * fact(j-1))
   for k in range(1,len(v)):
-    v[k] = fact(k-1) * sum([scale(k, j)* v[k-j] * u[j] for j in range(1, k+1)])
+    v[k] = fact(k-1) * sum([_scale(k, j)* v[k-j] * u[j] for j in range(1, k+1)])
   primal_out, *series_out = v
   return primal_out, series_out
 jet_rules[lax.exp_p] = _exp_taylor
 
+def _expm1_taylor(primals_in, series_in):
+  x, = primals_in
+  series, = series_in
+  u = [x] + series
+  v = [lax.exp(x)] + [None] * len(series)
+  for k in range(1,len(v)):
+    v[k] = fact(k-1) * sum([_scale(k, j)* v[k-j] * u[j] for j in range(1, k+1)])
+  primal_out, *series_out = v
+  return lax.expm1(x), series_out
+jet_rules[lax.expm1_p] = _expm1_taylor
+
 def _log_taylor(primals_in, series_in):
   x, = primals_in
   series, = series_in
   u = [x] + series
   v = [lax.log(x)] + [None] * len(series)
-  def scale(k, j): return 1. / (fact(k-j) * fact(j-1))
   for k in range(1, len(v)):
-    conv = sum([scale(k, j) * v[j] * u[k-j] for j in range(1, k)])
+    conv = sum([_scale(k, j) * v[j] * u[k-j] for j in range(1, k)])
     v[k] = (u[k] - fact(k - 1) * conv) / u[0]
   primal_out, *series_out = v
   return primal_out, series_out
 jet_rules[lax.log_p] = _log_taylor
 
+def _log1p_taylor(primals_in, series_in):
+  x, = primals_in
+  series, = series_in
+  u = [x + 1] + series
+  v = [lax.log(x + 1)] + [None] * len(series)
+  for k in range(1, len(v)):
+    conv = sum([_scale(k, j) * v[j] * u[k-j] for j in range(1, k)])
+    v[k] = (u[k] - fact(k - 1) * conv) / u[0]
+  primal_out, *series_out = v
+  return primal_out, series_out
+jet_rules[lax.log1p_p] = _log1p_taylor
+
 def _div_taylor_rule(primals_in, series_in, **params):
   x, y = primals_in
   x_terms, y_terms = series_in
@@ -223,22 +301,39 @@ def _gather_taylor_rule(primals_in, series_in, **params):
   return primal_out, series_out
 jet_rules[lax.gather_p] = _gather_taylor_rule
 
-def _reduce_max_taylor_rule(primals_in, series_in, **params):
-  operand, = primals_in
-  gs, = series_in
-  primal_out = lax.reduce_max_p.bind(operand, **params)
-  axes = params.pop("axes", None)
-  primal_dtype = gs[0].dtype
-  shape = [1 if i in axes else d for i, d in enumerate(operand.shape)]
-  location_indicators = lax.convert_element_type(
-        lax._eq_meet(operand, lax.reshape(primal_out, shape)), primal_dtype)
-  counts = lax._reduce_sum(location_indicators, axes)
-  def _reduce_chooser_taylor_rule(g):
-    return lax.div(lax._reduce_sum(lax.mul(g, location_indicators), axes), counts)
-  series_out = [_reduce_chooser_taylor_rule(g) for g in gs]
+def _gen_reduce_choose_taylor_rule(chooser_fun):
+  def chooser_taylor_rule(primals_in, series_in, **params):
+    operand, = primals_in
+    gs, = series_in
+    primal_out = chooser_fun(operand, **params)
+    axes = params.pop("axes", None)
+    primal_dtype = gs[0].dtype
+    shape = [1 if i in axes else d for i, d in enumerate(operand.shape)]
+    location_indicators = lax.convert_element_type(
+          lax._eq_meet(operand, lax.reshape(primal_out, shape)), primal_dtype)
+    counts = lax._reduce_sum(location_indicators, axes)
+    def _reduce_chooser_taylor_rule(g):
+      return lax.div(lax._reduce_sum(lax.mul(g, location_indicators), axes), counts)
+    series_out = [_reduce_chooser_taylor_rule(g) for g in gs]
+    return primal_out, series_out
+  return chooser_taylor_rule
+jet_rules[lax.reduce_max_p] = _gen_reduce_choose_taylor_rule(lax.reduce_max_p.bind)
+jet_rules[lax.reduce_min_p] = _gen_reduce_choose_taylor_rule(lax.reduce_min_p.bind)
+
+def _abs_taylor_rule(x, series_in, **params):
+  x, = x
+  primal_out = lax.abs_p.bind(x, **params)
+  negs = lax.select(lax.lt(x, 0.0), lax.full_like(x, -1), lax.full_like(x, 1.0))
+  fix_sign = lambda y: negs * y
+  series_out = [fix_sign(*terms_in, **params) for terms_in in zip(*series_in)]
   return primal_out, series_out
-jet_rules[lax.reduce_max_p] = _reduce_max_taylor_rule
+jet_rules[lax.abs_p] = _abs_taylor_rule
 
+def _select_taylor_rule(primal_in, series_in, **params):
+  b, x, y = primal_in
+  primal_out = lax.select_p.bind(b, x, y, **params)
+  sel = lambda _, x, y: lax.select(b, x, y)
+  series_out = [sel(*terms_in, **params) for terms_in in zip(*series_in)]
+  return primal_out, series_out
+jet_rules[lax.select_p] = _select_taylor_rule
 
-from jax.interpreters import xla
-deflinear(xla.device_put_p)
